"""Provides a client for the Halo Infinite API."""

from functools import cached_property

from aiohttp import ClientSession

from .services import (
    DiscoveryUgcService,
    GameCmsHacsService,
    ProfileService,
    SkillService,
    StatsService,
)

__all__ = ["HaloInfiniteClient"]


class HaloInfiniteClient:
    """A client for the Halo Infinite API."""

    def __init__(
        self,
        session: ClientSession,
        spartan_token: str,
        clearance_token: str,
        requests_per_second: int | None = 5,
    ) -> None:
        """Initialize a client for the Halo Infinite API.

        Raw `aiohttp.ClientResponses` are returned from each method. The
        caller is responsible for handling the response via custom parsing or by
        using one of the provided parsers from the `spnkr.parsers` module.

        Args:
            session: The aiohttp session to use.
            spartan_token: The spartan token used to authenticate with the API.
            clearance_token: The clearance token used to authenticate with the
                API.
            requests_per_second: The rate limit to use. Note that this rate
                limit is enforced per service, not globally. Defaults to 5
                requests per second. Set to None to disable rate limiting.
        """
        session.headers.clear()
        session.headers["Accept"] = "application/json"
        session.headers["x-343-authorization-spartan"] = spartan_token
        session.headers["343-clearance"] = clearance_token
        self._session = session
        self._requests_per_second = requests_per_second

    @cached_property
    def profile(self) -> ProfileService:
        """Profile data service. Get user data, such as XUIDs/gamertags."""
        return ProfileService(self._session, self._requests_per_second)

    @cached_property
    def gamecms_hacs(self) -> GameCmsHacsService:
        """Game content management data service (e.g., medal metadata)"""
        return GameCmsHacsService(self._session, self._requests_per_second)

    @cached_property
    def skill(self) -> SkillService:
        """Skill data service. Retrieve MMR and CSR data by match or playlist."""
        return SkillService(self._session, self._requests_per_second)

    @cached_property
    def stats(self) -> StatsService:
        """Stats data service. Retrieve match history and match stats."""
        return StatsService(self._session, self._requests_per_second)

    @cached_property
    def discovery_ugc(self) -> DiscoveryUgcService:
        """User-generated content discovery data service (maps, modes, etc.)."""
        return DiscoveryUgcService(self._session, self._requests_per_second)



"""Test the spnkr.tools module."""

from enum import IntEnum

import pytest

from spnkr import tools
from spnkr.parsers.refdata import SubTier, Tier

CSR_TIERS = [
    (0, Tier.BRONZE, SubTier.ONE),
    (299, Tier.BRONZE, SubTier.SIX),
    (300, Tier.SILVER, SubTier.ONE),
    (599, Tier.SILVER, SubTier.SIX),
    (600, Tier.GOLD, SubTier.ONE),
    (899, Tier.GOLD, SubTier.SIX),
    (900, Tier.PLATINUM, SubTier.ONE),
    (1199, Tier.PLATINUM, SubTier.SIX),
    (1200, Tier.DIAMOND, SubTier.ONE),
    (1499, Tier.DIAMOND, SubTier.SIX),
    (1500, Tier.ONYX, SubTier.ONE),
    (3000, Tier.ONYX, SubTier.ONE),
]


def test_ranking_str():
    """Test the __str__ method of the Rank class."""
    ranking = tools.Rank(Tier.BRONZE, SubTier.ONE)
    assert str(ranking) == "Bronze 1"


def test_ranking_str_onyx():
    """Test the __str__ method of the Rank class when the tier is Onyx."""
    ranking = tools.Rank(Tier.ONYX, SubTier.ONE)
    assert str(ranking) == "Onyx"


@pytest.mark.parametrize("csr,tier,sub_tier", CSR_TIERS)
def test_get_rank_from_csr(csr: float, tier: Tier, sub_tier: SubTier):
    """Test that the correct tier and sub-tier are returned."""
    assert tools.get_rank_from_csr(csr) == (tier, sub_tier)


def test_intenum_to_mapping():
    """Test the _intenum_to_mapping function."""

    class TestEnum(IntEnum):
        A = 1
        B = 2
        C = 3

    assert tools._intenum_to_mapping(TestEnum) == {1: "A", 2: "B", 3: "C"}



"""Parsing functions for the "stats" authority."""

import datetime as dt
from typing import Any, NamedTuple
from uuid import UUID

from ...xuid import wrap_xuid
from ..refdata import (
    BotDifficulty,
    GameVariantCategory,
    LifecycleMode,
    Outcome,
    PlayerType,
)


class MatchCountRecord(NamedTuple):
    """A player's match count summary.

    Attributes:
        player_id: The player's Xbox Live ID.
        total: The total number of matches played.
        custom: The number of custom matches played.
        matchmade: The number of matchmade matches played.
        local: The number of local matches played.
    """

    player_id: str
    total: int
    custom: int
    matchmade: int
    local: int


class MatchHistoryRecord(NamedTuple):
    """A single match from a player's match history.

    Attributes:
        match_id: The match GUID.
        start_time: The UTC datetime the match started.
        end_time: The UTC datetime the match ended.
        duration: The duration of the match.
        lifecycle_mode: The lifecycle mode of the match.
        game_variant_category: The game variant category of the match.
        level_id: The ID of the level the match was played on.
        map_asset_id: The asset ID of the map variant.
        map_version_id: The version ID of the map variant.
        game_variant_asset_id: The asset ID of the game mode.
        game_variant_version_id: The version ID of the game mode.
        playlist_asset_id: The asset ID of the playlist.
        playlist_version_id: The version ID of the playlist.
        map_mode_pair_asset_id: The asset ID of the map mode pair.
        map_mode_pair_version_id: The version ID of the map mode pair.
        season_id: The ID of the season the match was played in.
        playable_duration: The duration of the match that was playable.
        last_team_id: The ID of the team the player was on at the end of the match.
        outcome: The outcome of the match for the player.
        rank: The player's rank in the match.
        present_at_end_of_match: Whether the player was present at the end of the match.
    """

    match_id: UUID
    start_time: dt.datetime
    end_time: dt.datetime
    duration: dt.timedelta
    lifecycle_mode: LifecycleMode
    game_variant_category: GameVariantCategory
    level_id: str
    map_asset_id: str
    map_version_id: int
    game_variant_asset_id: str
    game_variant_version_id: int
    playlist_asset_id: str | None
    playlist_version_id: int | None
    map_mode_pair_asset_id: str | None
    map_mode_pair_version_id: int | None
    season_id: int | None
    playable_duration: dt.timedelta
    last_team_id: int
    outcome: Outcome
    rank: int
    present_at_end_of_match: bool


class MatchInfoRecord(NamedTuple):
    """Information about a single match.

    Attributes:
        match_id: The match GUID.
        start_time: The UTC datetime the match started.
        end_time: The UTC datetime the match ended.
        duration: The duration of the match.
        lifecycle_mode: The lifecycle mode of the match.
        game_variant_category: The game variant category of the match.
        level_id: The ID of the level the match was played on.
        map_asset_id: The asset ID of the map variant.
        map_version_id: The version ID of the map variant.
        game_variant_asset_id: The asset ID of the game mode.
        game_variant_version_id: The version ID of the game mode.
        playlist_asset_id: The asset ID of the playlist.
        playlist_version_id: The version ID of the playlist.
        map_mode_pair_asset_id: The asset ID of the map mode pair.
        map_mode_pair_version_id: The version ID of the map mode pair.
        season_id: The ID of the season the match was played in.
        playable_duration: The duration of the match that was playable.
    """

    match_id: UUID
    start_time: dt.datetime
    end_time: dt.datetime
    duration: dt.timedelta
    lifecycle_mode: LifecycleMode
    game_variant_category: GameVariantCategory
    level_id: str
    map_asset_id: str
    map_version_id: int
    game_variant_asset_id: str
    game_variant_version_id: int
    playlist_asset_id: str | None
    playlist_version_id: int | None
    map_mode_pair_asset_id: str | None
    map_mode_pair_version_id: int | None
    season_id: int | None
    playable_duration: dt.timedelta


class TeamCoreStatsRecord(NamedTuple):
    """A team's core stats for a single match.

    Attributes:
        match_id: The match GUID.
        team_id: The team ID.
        outcome: The outcome of the match for the team.
        rank: The team's rank in the match.
        score: The team's score.
        personal_score: The team's personal score.
        rounds_won: The number of rounds the team won.
        rounds_lost: The number of rounds the team lost.
        rounds_tied: The number of rounds the team tied.
        kills: The number of kills the team had.
        deaths: The number of times the team died.
        assists: The number of assists the team had.
        kda: The team's kill-death-assist metric (kills + assists / 3 - deaths).
        suicides: The number of times the team's members committed suicide.
        betrayals: The number of times the team's members killed each other.
        average_life_duration: The average duration of the team's lives.
        grenade_kills: The number of grenade kills the team had.
        headshot_kills: The number of headshot kills the team had.
        melee_kills: The number of melee kills the team had.
        power_weapon_kills: The number of power weapon kills the team had.
        shots_fired: The number of shots the team fired.
        shots_hit: The number of shots the team hit.
        accuracy: The team's accuracy (shots_hit / shots_fired * 100).
        damage_dealt: The amount of damage the team dealt.
        damage_taken: The amount of damage the team took.
        callout_assists: The number of marking assists the team had.
        vehicle_destroys: The number of vehicles the team destroyed.
        driver_assists: The number of driver assists the team had.
        hijacks: The number of vehicles the team hijacked.
        emp_assists: The number of EMP assists the team had.
        max_killing_spree: The maximum killing spree across all team members.
        spawns: The number of times the team spawned.
    """

    match_id: UUID
    team_id: int
    outcome: Outcome
    rank: int
    score: int
    personal_score: int
    rounds_won: int
    rounds_lost: int
    rounds_tied: int
    kills: int
    deaths: int
    assists: int
    kda: float
    suicides: int
    betrayals: int
    average_life_duration: dt.timedelta
    grenade_kills: int
    headshot_kills: int
    melee_kills: int
    power_weapon_kills: int
    shots_fired: int
    shots_hit: int
    accuracy: float
    damage_dealt: int
    damage_taken: int
    callout_assists: int
    vehicle_destroys: int
    driver_assists: int
    hijacks: int
    emp_assists: int
    max_killing_spree: int
    spawns: int


class PlayerCoreStatsRecord(NamedTuple):
    """A player's core stats for a single match.

    Attributes:
        match_id: The match GUID.
        player_id: The player's Xbox Live ID.
        player_type: The player's type.
        bot_difficulty: The bot's difficulty, if the player is a bot.
        last_team_id: The ID of the team the player was on at the end of the match.
        outcome: The outcome of the match for the player.
        rank: The player's rank in the match.
        present_at_beginning: Whether the player was present at the beginning of the match.
        present_at_completion: Whether the player was present at the completion of the match.
        time_played: The amount of time the player was present in the match.
        team_id: The ID of the team the player was on.
        score: The player's score.
        personal_score: The player's personal score.
        rounds_won: The number of rounds the player's team won.
        rounds_lost: The number of rounds the player's team lost.
        rounds_tied: The number of rounds the player's team tied.
        kills: The number of kills the player had.
        deaths: The number of times the player died.
        assists: The number of assists the player had.
        kda: The player's kill-death-assist metric (kills + assists / 3 - deaths).
        suicides: The number of times the player committed suicide.
        betrayals: The number of times the player betrayed a teammate.
        average_life_duration: The average duration of the player's lives.
        grenade_kills: The number of grenade kills the player had.
        headshot_kills: The number of headshot kills the player had.
        melee_kills: The number of melee kills the player had.
        power_weapon_kills: The number of power weapon kills the player had.
        shots_fired: The number of shots the player fired.
        shots_hit: The number of shots the player hit.
        accuracy: The player's accuracy (shots_hit / shots_fired * 100).
        damage_dealt: The amount of damage the player dealt.
        damage_taken: The amount of damage the player took.
        callout_assists: The number of callout assists the player had.
        vehicle_destroys: The number of vehicles the player destroyed.
        driver_assists: The number of driver assists the player had.
        hijacks: The number of vehicles the player hijacked.
        emp_assists: The number of EMP assists the player had.
        max_killing_spree: The maximum killing spree the player had.
        spawns: The number of times the player spawned.
    """

    match_id: UUID
    player_id: str
    player_type: PlayerType
    bot_difficulty: BotDifficulty | None
    last_team_id: int
    outcome: Outcome
    rank: int
    present_at_beginning: bool
    present_at_completion: bool
    time_played: dt.timedelta
    team_id: int
    score: int
    personal_score: int
    rounds_won: int
    rounds_lost: int
    rounds_tied: int
    kills: int
    deaths: int
    assists: int
    kda: float
    suicides: int
    betrayals: int
    average_life_duration: dt.timedelta
    grenade_kills: int
    headshot_kills: int
    melee_kills: int
    power_weapon_kills: int
    shots_fired: int
    shots_hit: int
    accuracy: float
    damage_dealt: int
    damage_taken: int
    callout_assists: int
    vehicle_destroys: int
    driver_assists: int
    hijacks: int
    emp_assists: int
    max_killing_spree: int
    spawns: int


class PlayerMedalRecord(NamedTuple):
    """A single medal earned by a player in a match.

    Attributes:
        match_id: The match GUID.
        player_id: The player's Xbox Live ID.
        team_id: The team ID of the player.
        name_id: The medal's ID.
        count: The number of times the medal was earned.
    """

    match_id: UUID
    player_id: str
    team_id: int
    name_id: int
    count: int


class PlayerServiceRecord(NamedTuple):
    """A player's service record in a given context.

    Attributes:
        player_id: The player's Xbox Live ID.
        time_played: The amount of time the player was present in the match.
        matches_completed: The number of matches the player completed.
        wins: The number of matches the player won.
        losses: The number of matches the player lost.
        ties: The number of matches the player tied.
        score: The player's score.
        personal_score: The player's personal score.
        rounds_won: The number of rounds the player's team won.
        rounds_lost: The number of rounds the player's team lost.
        rounds_tied: The number of rounds the player's team tied.
        kills: The number of kills the player had.
        deaths: The number of times the player died.
        assists: The number of assists the player had.
        average_kda: The player's kill-death-assist metric (kills + assists / 3 - deaths).
        suicides: The number of times the player committed suicide.
        betrayals: The number of times the player betrayed a teammate.
        grenade_kills: The number of grenade kills the player had.
        headshot_kills: The number of headshot kills the player had.
        melee_kills: The number of melee kills the player had.
        power_weapon_kills: The number of power weapon kills the player had.
        shots_fired: The number of shots the player fired.
        shots_hit: The number of shots the player hit.
        accuracy: The player's accuracy (shots_hit / shots_fired * 100).
        damage_dealt: The amount of damage the player dealt.
        damage_taken: The amount of damage the player took.
        callout_assists: The number of callout assists the player had.
        vehicle_destroys: The number of vehicles the player destroyed.
        driver_assists: The number of driver assists the player had.
        hijacks: The number of vehicles the player hijacked.
        emp_assists: The number of EMP assists the player had.
        max_killing_spree: The maximum killing spree the player had.
        spawns: The number of times the player spawned.
    """

    player_id: str
    time_played: dt.timedelta
    matches_completed: int
    wins: int
    losses: int
    ties: int
    score: int
    personal_score: int
    rounds_won: int
    rounds_lost: int
    rounds_tied: int
    kills: int
    deaths: int
    assists: int
    average_kda: float
    suicides: int
    betrayals: int
    grenade_kills: int
    headshot_kills: int
    melee_kills: int
    power_weapon_kills: int
    shots_fired: int
    shots_hit: int
    accuracy: float
    damage_dealt: int
    damage_taken: int
    callout_assists: int
    vehicle_destroys: int
    driver_assists: int
    hijacks: int
    emp_assists: int
    max_killing_spree: int
    spawns: int


def parse_match_count(
    xuid: str | int, match_count: dict[str, Any]
) -> MatchCountRecord:
    """Parse a match count response into a match count record.

    Args:
        xuid: The player's Xbox Live ID.
        match_count: The deserialized JSON from the client's `get_match_count`
            method.

    Returns:
        A match count record.
    """
    return MatchCountRecord(
        player_id=wrap_xuid(xuid),
        total=match_count["MatchesPlayedCount"],
        custom=match_count["CustomMatchesPlayedCount"],
        matchmade=match_count["MatchmadeMatchesPlayedCount"],
        local=match_count["LocalMatchesPlayedCount"],
    )


def parse_match_history(
    match_history: dict[str, Any]
) -> list[MatchHistoryRecord]:
    """Parse a match history response into a list of match history records.

    Args:
        match_history: The deserialized JSON from the client's
            `get_match_history` method.

    Returns:
        A list of match history records.
    """
    return [_parse_match_history_result(r) for r in match_history["Results"]]


def parse_team_core_stats(
    match_stats: dict[str, Any]
) -> list[TeamCoreStatsRecord]:
    """Parse a match stats response into a list of team core stats records.

    Args:
        match_stats: The deserialized JSON from the client's `get_match_stats`
            method.

    Returns:
        A list of team core stats records.
    """
    out: list[TeamCoreStatsRecord] = []
    for team in match_stats["Teams"]:
        entry = TeamCoreStatsRecord(
            match_id=match_stats["MatchId"],
            team_id=team["TeamId"],
            outcome=Outcome(team["Outcome"]),
            rank=team["Rank"],
            **_parse_core_stats(team["Stats"]["CoreStats"]),
        )
        out.append(entry)
    return out


def parse_match_info(match_stats: dict[str, Any]) -> MatchInfoRecord:
    """Parse a match stats response into a match info record.

    Args:
        match_stats: The deserialized JSON from the client's `get_match_stats`
            method.

    Returns:
        A match info record.
    """
    return MatchInfoRecord(
        match_id=match_stats["MatchId"],
        **_parse_match_info(match_stats["MatchInfo"]),
    )


def parse_player_core_stats(
    match_stats: dict[str, Any]
) -> list[PlayerCoreStatsRecord]:
    """Parse a match stats response into a list of player core stats records.

    Args:
        match_stats: The deserialized JSON from the client's `get_match_stats`
            method.

    Returns:
        A list of player core stats records.
    """
    out: list[PlayerCoreStatsRecord] = []
    for player in match_stats["Players"]:
        participation = player["ParticipationInfo"]
        bot_attributes = player["BotAttributes"]
        difficulty = None
        if bot_attributes is not None:
            difficulty = BotDifficulty(bot_attributes["Difficulty"])
        for team in player["PlayerTeamStats"]:
            entry = PlayerCoreStatsRecord(
                match_id=match_stats["MatchId"],
                player_id=player["PlayerId"],
                player_type=PlayerType(player["PlayerType"]),
                bot_difficulty=difficulty,
                last_team_id=player["LastTeamId"],
                outcome=Outcome(player["Outcome"]),
                rank=player["Rank"],
                present_at_beginning=participation["PresentAtBeginning"],
                present_at_completion=participation["PresentAtCompletion"],
                time_played=_parse_iso_duration(participation["TimePlayed"]),
                team_id=team["TeamId"],
                **_parse_core_stats(team["Stats"]["CoreStats"]),
            )
            out.append(entry)
    return out


def parse_player_medals(match_stats: dict[str, Any]) -> list[PlayerMedalRecord]:
    """Parse a match stats response into a list of player medal records.

    Args:
        match_stats: The deserialized JSON from the client's `get_match_stats`
            method.

    Returns:
        A list of player medal records.
    """
    out: list[PlayerMedalRecord] = []
    for player in match_stats["Players"]:
        for team in player["PlayerTeamStats"]:
            for medal in team["Stats"]["CoreStats"]["Medals"]:
                entry = PlayerMedalRecord(
                    match_id=match_stats["MatchId"],
                    player_id=player["PlayerId"],
                    team_id=team["TeamId"],
                    name_id=medal["NameId"],
                    count=medal["Count"],
                )
                out.append(entry)
    return out


def parse_service_record(
    xuid: str | int, service_record: dict[str, Any]
) -> PlayerServiceRecord:
    """Parse a service record response into a player service record.

    Args:
        xuid: The player's Xbox Live ID.
        service_record: The deserialized JSON from the client's
            `get_service_record` method.

    Returns:
        A player service record.
    """
    stats = service_record["CoreStats"]
    return PlayerServiceRecord(
        player_id=wrap_xuid(xuid),
        time_played=_parse_iso_duration(service_record["TimePlayed"]),
        matches_completed=service_record["MatchesCompleted"],
        wins=service_record["Wins"],
        losses=service_record["Losses"],
        ties=service_record["Ties"],
        score=stats["Score"],
        personal_score=stats["PersonalScore"],
        rounds_won=stats["RoundsWon"],
        rounds_lost=stats["RoundsLost"],
        rounds_tied=stats["RoundsTied"],
        kills=stats["Kills"],
        deaths=stats["Deaths"],
        assists=stats["Assists"],
        average_kda=stats["AverageKDA"],
        suicides=stats["Suicides"],
        betrayals=stats["Betrayals"],
        grenade_kills=stats["GrenadeKills"],
        headshot_kills=stats["HeadshotKills"],
        melee_kills=stats["MeleeKills"],
        power_weapon_kills=stats["PowerWeaponKills"],
        shots_fired=stats["ShotsFired"],
        shots_hit=stats["ShotsHit"],
        accuracy=stats["Accuracy"],
        damage_dealt=stats["DamageDealt"],
        damage_taken=stats["DamageTaken"],
        callout_assists=stats["CalloutAssists"],
        vehicle_destroys=stats["VehicleDestroys"],
        driver_assists=stats["DriverAssists"],
        hijacks=stats["Hijacks"],
        emp_assists=stats["EmpAssists"],
        max_killing_spree=stats["MaxKillingSpree"],
        spawns=stats["Spawns"],
    )


def _parse_match_history_result(result: dict[str, Any]) -> MatchHistoryRecord:
    info = result["MatchInfo"]
    return MatchHistoryRecord(
        match_id=result["MatchId"],
        **_parse_match_info(info),
        last_team_id=result["LastTeamId"],
        outcome=Outcome(result["Outcome"]),
        rank=result["Rank"],
        present_at_end_of_match=result["PresentAtEndOfMatch"],
    )


def _parse_match_info(info: dict[str, Any]) -> dict[str, Any]:
    playlist = info["Playlist"]
    playlist_asset_id = playlist_version_id = None
    if playlist is not None:
        playlist_asset_id = playlist["AssetId"]
        playlist_version_id = playlist["VersionId"]
    map_mode = info["PlaylistMapModePair"]
    map_mode_asset_id = map_mode_version_id = None
    if map_mode is not None:
        map_mode_asset_id = map_mode["AssetId"]
        map_mode_version_id = map_mode["VersionId"]
    return {
        "start_time": dt.datetime.fromisoformat(info["StartTime"]),
        "end_time": dt.datetime.fromisoformat(info["EndTime"]),
        "duration": _parse_iso_duration(info["Duration"]),
        "lifecycle_mode": LifecycleMode(info["LifecycleMode"]),
        "game_variant_category": GameVariantCategory(
            info["GameVariantCategory"]
        ),
        "level_id": info["LevelId"],
        "map_asset_id": info["MapVariant"]["AssetId"],
        "map_version_id": info["MapVariant"]["VersionId"],
        "game_variant_asset_id": info["UgcGameVariant"]["AssetId"],
        "game_variant_version_id": info["UgcGameVariant"]["VersionId"],
        "playlist_asset_id": playlist_asset_id,
        "playlist_version_id": playlist_version_id,
        "map_mode_pair_asset_id": map_mode_asset_id,
        "map_mode_pair_version_id": map_mode_version_id,
        "season_id": info.get("SeasonId"),
        "playable_duration": _parse_iso_duration(info["PlayableDuration"]),
    }


def _parse_core_stats(stats: dict[str, Any]) -> dict[str, Any]:
    return {
        "score": stats["Score"],
        "personal_score": stats["PersonalScore"],
        "rounds_won": stats["RoundsWon"],
        "rounds_lost": stats["RoundsLost"],
        "rounds_tied": stats["RoundsTied"],
        "kills": stats["Kills"],
        "deaths": stats["Deaths"],
        "assists": stats["Assists"],
        "kda": stats["KDA"],
        "suicides": stats["Suicides"],
        "betrayals": stats["Betrayals"],
        "average_life_duration": _parse_iso_duration(
            stats["AverageLifeDuration"]
        ),
        "grenade_kills": stats["GrenadeKills"],
        "headshot_kills": stats["HeadshotKills"],
        "melee_kills": stats["MeleeKills"],
        "power_weapon_kills": stats["PowerWeaponKills"],
        "shots_fired": stats["ShotsFired"],
        "shots_hit": stats["ShotsHit"],
        "accuracy": stats["Accuracy"],
        "damage_dealt": stats["DamageDealt"],
        "damage_taken": stats["DamageTaken"],
        "callout_assists": stats["CalloutAssists"],
        "vehicle_destroys": stats["VehicleDestroys"],
        "driver_assists": stats["DriverAssists"],
        "hijacks": stats["Hijacks"],
        "emp_assists": stats["EmpAssists"],
        "max_killing_spree": stats["MaxKillingSpree"],
        "spawns": stats["Spawns"],
    }


def _iso_split(value: str, separator: str) -> tuple[float, str]:
    """Split an ISO 8601 duration string at a given separator.

    Return the count and the remaining string.
    """
    count = None
    if separator in value:
        count, value = value.split(separator)
    return (float(count or 0), value)


def _parse_iso_duration(value: str) -> dt.timedelta:
    """Parse an ISO 8601 duration string to a timedelta object.

    Example: PT12M34.1102934S => 12 minutes, 34.1102934 seconds

    Only days, hours, minutes, and seconds are supported.
    """
    haystack = value.split("P")[-1]  # Remove leading P
    days, haystack = _iso_split(haystack, "D")
    _, haystack = _iso_split(haystack, "T")
    hours, haystack = _iso_split(haystack, "H")
    minutes, haystack = _iso_split(haystack, "M")
    seconds, haystack = _iso_split(haystack, "S")
    return dt.timedelta(
        days=days, hours=hours, minutes=minutes, seconds=seconds
    )



"""Custom exceptions for spnkr."""


class InvalidXuidError(ValueError):
    """Raised when an invalid XUID is provided."""

    def __init__(self, xuid: str | int) -> None:
        super().__init__(f"Invalid XUID: {xuid!r}")



"""Test StatsService."""

import pytest

from spnkr.services.stats import StatsService


@pytest.fixture
def service(session):
    return StatsService(session)


@pytest.mark.asyncio
async def test_get_match_count(session, service: StatsService):
    await service.get_match_count(1234567890123456)
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/matches/count"
    )


@pytest.mark.asyncio
async def test_get_service_record(session, service: StatsService):
    await service.get_service_record(1234567890123456)
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/matchmade/servicerecord",
        params={},
    )


@pytest.mark.asyncio
async def test_get_service_record_season_playlist(
    session, service: StatsService
):
    await service.get_service_record(
        1234567890123456,
        season_id="season_id",
        playlist_asset_id="playlist_asset_id",
    )
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/matchmade/servicerecord",
        params={
            "seasonid": "season_id",
            "playlistassetid": "playlist_asset_id",
        },
    )


@pytest.mark.asyncio
async def test_get_service_record_game_variant_category_ranked(
    session, service: StatsService
):
    await service.get_service_record(
        1234567890123456, game_variant_category=6, is_ranked=True
    )
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/matchmade/servicerecord",
        params={"gamevariantcategory": "6", "isranked": "True"},
    )


@pytest.mark.asyncio
async def test_get_service_record_invalid_match_type(service: StatsService):
    with pytest.raises(ValueError):
        await service.get_service_record(1234567890123456, match_type="invalid")  # type: ignore


@pytest.mark.asyncio
async def test_get_service_record_invalid_filter_combination(
    service: StatsService,
):
    with pytest.raises(ValueError):
        await service.get_service_record(1234567890123456, playlist_asset_id="")


@pytest.mark.asyncio
async def test_get_service_record_non_matchmade_params_ignored(
    session,
    service: StatsService,
):
    await service.get_service_record(
        1234567890123456, match_type="custom", is_ranked=True
    )
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/custom/servicerecord",
        params={},
    )


@pytest.mark.asyncio
async def test_get_match_history(session, service: StatsService):
    await service.get_match_history(1234567890123456, 0, 10, "all")
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/players/xuid(1234567890123456)/matches",
        params={"start": 0, "count": 10, "type": "all"},
    )


@pytest.mark.asyncio
async def test_get_match_stats(session, service: StatsService):
    await service.get_match_stats("match_id")
    session.get.assert_called_with(
        "https://halostats.svc.halowaypoint.com:443/hi/matches/match_id/stats"
    )



"""Test the spnkr.auth.oauth module."""

import json
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock

import pytest
from aiohttp import ClientSession

from spnkr.auth import app, oauth

RESPONSES = Path("tests/responses")


class MockResponse:
    def __init__(self, json_data: Any):
        self.json_data = json_data

    async def json(self) -> Any:
        return self.json_data


class MockSession:
    def __init__(self, response: Any):
        self.response = response

    async def get(self, *args, **kwargs) -> Any:
        return self.response

    async def post(self, *args, **kwargs) -> Any:
        return self.response


@pytest.fixture
def azure_app():
    return app.AzureApp("client_id", "client_secret", "redirect_uri")


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


def test_oauth2_token_access_token():
    """Test that the access token value is returned."""
    response = load_response("oauth2")
    oauth_token = oauth.OAuth2Token(response)
    assert oauth_token.access_token == "abcdefg"


def test_oauth2_token_refresh_token():
    """Test that the refresh token value is returned."""
    response = load_response("oauth2")
    oauth_token = oauth.OAuth2Token(response)
    assert oauth_token.refresh_token == "hijklmnop"


def test_generate_authorization_url(azure_app):
    """Test that the authorization URL is generated."""
    expected = (
        "https://login.live.com/oauth20_authorize.srf?"
        "client_id=client_id&"
        "response_type=code&"
        "approval_prompt=auto&"
        "scope=Xboxlive.signin+Xboxlive.offline_access&"
        "redirect_uri=redirect_uri"
    )
    assert oauth.generate_authorization_url(azure_app) == expected


@pytest.mark.asyncio
async def test_request_oauth_token(azure_app, monkeypatch):
    mock = AsyncMock()
    monkeypatch.setattr(oauth, "_oauth2_token_request", mock)
    expected_data = {
        "grant_type": "authorization_code",
        "code": "auth_code",
        "scope": "Xboxlive.signin Xboxlive.offline_access",
        "redirect_uri": "redirect_uri",
    }
    async with ClientSession() as session:
        await oauth.request_oauth_token(session, "auth_code", azure_app)

    mock.assert_called_with(session, expected_data, azure_app)


@pytest.mark.asyncio
async def test_refresh_oauth_token(azure_app, monkeypatch):
    mock = AsyncMock()
    monkeypatch.setattr(oauth, "_oauth2_token_request", mock)
    expected_data = {
        "grant_type": "refresh_token",
        "scope": "Xboxlive.signin Xboxlive.offline_access",
        "refresh_token": "rt",
    }
    async with ClientSession() as session:
        await oauth.refresh_oauth_token(session, "rt", azure_app)

    mock.assert_called_with(session, expected_data, azure_app)


@pytest.mark.asyncio
async def test_oauth_token_request(azure_app):
    response = load_response("oauth2")
    session = MockSession(MockResponse(response))
    token = await oauth._oauth2_token_request(session, {}, azure_app)  # type: ignore
    assert token.access_token == "abcdefg"



"""User-generated content discovery data services."""

from uuid import UUID

from aiohttp import ClientResponse

from .base import BaseService

_HOST = "https://discovery-infiniteugc.svc.halowaypoint.com:443"


class DiscoveryUgcService(BaseService):
    """User-generated content discovery data services."""

    async def _get_asset(
        self, asset_type: str, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        url = f"{_HOST}/hi/{asset_type}/{asset_id}/versions/{version_id}"
        return await self._get(url)

    async def get_ugc_game_variant(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a game mode.

        Args:
            asset_id: The asset ID of the game variant.
            version_id: The version ID of the game variant.

        Parsers:
            - [UgcGameVariant][spnkr.parsers.pydantic.ugc_discovery.UgcGameVariant]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The game variant details.
        """
        return await self._get_asset("ugcGameVariants", asset_id, version_id)

    async def get_map_mode_pair(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a map mode pair.

        Args:
            asset_id: The asset ID of the map mode pair.
            version_id: The version ID of the map mode pair.

        Parsers:
            - [MapModePair][spnkr.parsers.pydantic.ugc_discovery.MapModePair]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The map mode pair details.
        """
        return await self._get_asset("mapModePairs", asset_id, version_id)

    async def get_map(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a map.

        Args:
            asset_id: The asset ID of the map.
            version_id: The version ID of the map.

        Parsers:
            - [Map][spnkr.parsers.pydantic.ugc_discovery.Map]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The map details.
        """
        return await self._get_asset("maps", asset_id, version_id)

    async def get_playlist(
        self, asset_id: str | UUID, version_id: str | UUID
    ) -> ClientResponse:
        """Get details about a playlist.

        Args:
            asset_id: The asset ID of the playlist.
            version_id: The version ID of the playlist.

        Parsers:
            - [Playlist][spnkr.parsers.pydantic.ugc_discovery.Playlist]
            - [parse_asset][spnkr.parsers.records.ugc_discovery.parse_asset]

        Returns:
            The playlist details.
        """
        return await self._get_asset("playlists", asset_id, version_id)



"""Models for the "gamecms_hacs" authority."""

from pydantic import Field

from ..refdata import MedalDifficulty, MedalType
from .base import CamelCaseModel


class TranslatableString(CamelCaseModel):
    """A string value accompanied by a dictionary of translations.

    Attributes:
        value: The string value.
        translations: A dictionary of language codes to translations for the string value.
    """

    value: str
    translations: dict[str, str]


class Medal(CamelCaseModel):
    """Metadata for a single medal.

    Attributes:
        name_id: The medal's ID.
        name: The medal's name.
        description: The medal's description.
        sprite_index: The index of the medal's sprite in the sprite sheet.
        sorting_weight: The medal's sorting weight.
        difficulty: The medal's difficulty, such as "normal" or "legendary".
        type: The medal's type, such as "mode" or "proficiency".
        personal_score: The amount of personal score awarded by obtaining the medal.
    """

    name_id: int
    name: TranslatableString
    description: TranslatableString
    sprite_index: int
    sorting_weight: int
    difficulty: MedalDifficulty = Field(alias="difficultyIndex")
    type: MedalType = Field(alias="typeIndex")
    personal_score: int


class SpriteSheet(CamelCaseModel):
    """Information about a sprite sheet that contains medal sprites.

    Attributes:
        path: The path to the sprite sheet, relative to the "gamecms_hacs" host.
        columns: The number of columns in the sprite sheet.
        size: The size of each sprite in the sprite sheet.
    """

    path: str
    columns: int
    size: int


class SpriteSheets(CamelCaseModel):
    """Sizes of sprite sheets that contain medal sprites.

    Attributes:
        small: A sprite sheet with small sprites.
        medium: A sprite sheet with medium sprites.
        extra_large: A sprite sheet with extra large sprites.
    """

    small: SpriteSheet
    medium: SpriteSheet
    extra_large: SpriteSheet = Field(alias="extra-large")


class MedalMetadata(CamelCaseModel):
    """Metadata for all medals.

    Attributes:
        difficulties: The list of possible difficulties.
        types: The list of possible types.
        sprites: Information about the medal sprite sheets that contain medal.
        medals: The list of available medals.
    """

    difficulties: list[str]
    types: list[str]
    sprites: SpriteSheets
    medals: list[Medal]



"""Test the spnkr.auth.xbox module."""

import json
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock

import pytest

from spnkr.auth import xbox

RESPONSES = Path("tests/responses")


class MockResponse:
    def __init__(self, json_data: Any):
        self.json_data = json_data

    async def json(self) -> Any:
        return self.json_data


class MockSession:
    def __init__(self, response: Any):
        self.post = AsyncMock(return_value=response)


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


def test_xau_response_token():
    """Test that the user token value is returned."""
    response = load_response("xbox_user")
    user_token = xbox.XAUResponse(response)
    assert user_token.token == "abcdefg"


def test_xsts_response_token():
    """Test that the xsts token is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert xsts_token.token == "123456789"


def test_xsts_response_xui():
    """Test that the xbox user info is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert len(xsts_token.xui) == 7


def test_xsts_response_xuid():
    """Test that the xsts player id is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert xsts_token.xuid == "xuid(2669321029139235)"


def test_xsts_response_userhash():
    """Test that the xsts user hash is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert xsts_token.userhash == "abcdefg"


def test_xsts_response_gamertag():
    """Test that the xsts gamertag is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert xsts_token.gamertag == "e"


def test_xsts_response_authorization_header_value():
    """Test that the xsts gamertag is returned."""
    response = load_response("xsts")
    xsts_token = xbox.XSTSResponse(response)
    assert xsts_token.authorization_header_value == "XBL3.0 x=abcdefg;123456789"


@pytest.mark.asyncio
async def test_request_user_token():
    response = load_response("xbox_user")
    session = MockSession(MockResponse(response))
    token = await xbox.request_user_token(session, "test")  # type: ignore
    assert token.token == "abcdefg"


@pytest.mark.asyncio
async def test_request_user_token_called_with():
    session = MockSession(MockResponse({}))
    await xbox.request_user_token(session, "test")  # type: ignore
    session.post.assert_called_with(
        "https://user.auth.xboxlive.com/user/authenticate",
        json={
            "RelyingParty": "http://auth.xboxlive.com",
            "TokenType": "JWT",
            "Properties": {
                "AuthMethod": "RPS",
                "SiteName": "user.auth.xboxlive.com",
                "RpsTicket": "d=test",
            },
        },
        headers={"x-xbl-contract-version": "1"},
    )


@pytest.mark.asyncio
async def test_request_xsts_token():
    response = load_response("xsts")
    session = MockSession(MockResponse(response))
    token = await xbox.request_xsts_token(session, "test", "party")  # type: ignore
    assert token.token == "123456789"


@pytest.mark.asyncio
async def test_request_xsts_token_called_with():
    session = MockSession(MockResponse({}))
    await xbox.request_xsts_token(session, "test", "party")  # type: ignore
    session.post.assert_called_with(
        "https://xsts.auth.xboxlive.com/xsts/authorize",
        json={
            "RelyingParty": "party",
            "TokenType": "JWT",
            "Properties": {
                "UserTokens": ["test"],
                "SandboxId": "RETAIL",
            },
        },
        headers={"x-xbl-contract-version": "1"},
    )



"""Test the spnkr.auth.halo module."""

import datetime as dt
import json
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock

import pytest

from spnkr.auth import halo

RESPONSES = Path("tests/responses")


class MockResponse:
    def __init__(self, json_data: Any):
        self.json_data = json_data

    async def json(self) -> Any:
        return self.json_data


class MockSession:
    def __init__(self, response: Any):
        self.get = AsyncMock(return_value=response)
        self.post = AsyncMock(return_value=response)


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


def test_spartan_token_token():
    """Test that the spartan token is returned."""
    response = load_response("spartan")
    spartan_token = halo.SpartanToken(response)
    assert spartan_token.token == "abcdef"


def test_spartan_token_expires_at():
    """Test that the expiration datetime is returned."""
    response = load_response("spartan")
    spartan_token = halo.SpartanToken(response)
    expected = dt.datetime(2999, 1, 27, 5, 3, 47, tzinfo=dt.timezone.utc)
    assert spartan_token.expires_at == expected


def test_clearance_token_token():
    """Test that the clearance token is returned."""
    response = load_response("clearance")
    clearance_token = halo.ClearanceToken(response)
    assert clearance_token.token == "xyz"


@pytest.mark.asyncio
async def test_request_spartan_token():
    """Test that a spartan token is returned."""
    response = load_response("spartan")
    session = MockSession(MockResponse(response))
    token = await halo.request_spartan_token(session, "test")  # type: ignore
    assert token.token == "abcdef"


@pytest.mark.asyncio
async def test_request_spartan_token_called_with():
    session = MockSession(MockResponse({}))
    await halo.request_spartan_token(session, "test")  # type: ignore
    session.post.assert_called_with(
        "https://settings.svc.halowaypoint.com/spartan-token",
        json={
            "Audience": "urn:343:s3:services",
            "MinVersion": "4",
            "Proof": [
                {
                    "Token": "test",
                    "TokenType": "Xbox_XSTSv3",
                }
            ],
        },
        headers={"Accept": "application/json"},
    )


@pytest.mark.asyncio
async def test_request_clearance_token():
    """Test that a clearance token is returned."""
    response = load_response("clearance")
    session = MockSession(MockResponse(response))
    token = await halo.request_clearance_token(session, "test")  # type: ignore
    assert token.token == "xyz"


@pytest.mark.asyncio
async def test_request_clearance_token_called_with():
    session = MockSession(MockResponse({}))
    await halo.request_clearance_token(session, "test")  # type: ignore
    session.get.assert_called_with(
        "https://settings.svc.halowaypoint.com/oban/flight-configurations/titles/hi/audiences/RETAIL/active",
        params={"sandbox": "UNUSED", "build": "222249.22.06.08.1730-0"},
        headers={"x-343-authorization-spartan": "test"},
    )



"""Skill data services."""

from typing import Iterable
from uuid import UUID

from aiohttp import ClientResponse

from ..xuid import wrap_xuid
from .base import BaseService

_HOST = "https://skill.svc.halowaypoint.com:443"


class SkillService(BaseService):
    """Skill data services."""

    async def get_match_skill(
        self, match_id: str | UUID, xuids: Iterable[str | int]
    ) -> ClientResponse:
        """Get player CSR and team MMR values for a given match and player list.

        Args:
            match_id: Halo Infinite match ID.
            xuids: The Xbox Live IDs of the match's players. Only
                players in this list will have their skill data returned.

        Parsers:
            - [MatchSkill][spnkr.parsers.pydantic.skill.MatchSkill]
            - [parse_match_skill][spnkr.parsers.records.skill.parse_match_skill]

        Returns:
            The skill data for the match.
        """
        url = f"{_HOST}/hi/matches/{match_id}/skill"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        return await self._get(url, params=params)

    async def get_playlist_csr(
        self, playlist_id: str | UUID, xuids: Iterable[str | int]
    ) -> ClientResponse:
        """Get player CSR values for a given playlist and player list.

        Args:
            playlist_id: Halo Infinite playlist asset ID.
            xuids: The Xbox Live IDs of the players.

        Parsers:
            - [PlaylistCsr][spnkr.parsers.pydantic.skill.PlaylistCsr]
            - [parse_playlist_csr][spnkr.parsers.records.skill.parse_playlist_csr]

        Returns:
            The summary CSR data for the players in the given playlist.
        """
        url = f"{_HOST}/hi/playlist/{playlist_id}/csrs"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        return await self._get(url, params=params)



"""Parse responses into pydantic models."""

from .gamecms_hacs import MedalMetadata
from .profile import User
from .skill import MatchSkill, PlaylistCsr
from .stats import MatchCount, MatchHistory, MatchStats, ServiceRecord
from .ugc_discovery import Map, MapModePair, Playlist, UgcGameVariant

__all__ = [
    "Map",
    "MapModePair",
    "MatchCount",
    "MatchHistory",
    "MatchSkill",
    "MatchStats",
    "MedalMetadata",
    "Playlist",
    "PlaylistCsr",
    "ServiceRecord",
    "UgcGameVariant",
    "User",
]



"""Profile data services."""

from typing import Iterable

from aiohttp import ClientResponse

from ..xuid import unwrap_xuid, wrap_xuid
from .base import BaseService

_HOST = "https://profile.svc.halowaypoint.com"


class ProfileService(BaseService):
    """Profile data services."""

    async def _get_user(self, user: str) -> ClientResponse:
        return await self._get(f"{_HOST}/users/{user}")

    async def get_current_user(self) -> ClientResponse:
        """Get the current user profile.

        Parsers:
            - [User][spnkr.parsers.pydantic.profile.User]
            - [parse_user][spnkr.parsers.records.profile.parse_user]

        Returns:
            The user.
        """
        return await self._get_user("me")

    async def get_user_by_gamertag(self, gamertag: str) -> ClientResponse:
        """Get user profile for the given gamertag.

        Args:
            gamertag: The gamertag of the player.

        Parsers:
            - [User][spnkr.parsers.pydantic.profile.User]
            - [parse_user][spnkr.parsers.records.profile.parse_user]

        Returns:
            The user.
        """
        return await self._get_user(f"gt({gamertag})")

    async def get_user_by_id(self, xuid: str | int) -> ClientResponse:
        """Get user profile for the given Xbox Live ID.

        Args:
            xuid: The Xbox Live ID of the player.

        Parsers:
            - [User][spnkr.parsers.pydantic.profile.User]
            - [parse_user][spnkr.parsers.records.profile.parse_user]

        Returns:
            The user.
        """
        return await self._get_user(wrap_xuid(xuid))

    async def get_users_by_id(
        self, xuids: Iterable[str | int]
    ) -> ClientResponse:
        """Get user profiles for the given list of Xbox Live IDs.

        Note that the JSON response is an array. This differs from the other
        endpoints, which return a single JSON object.

        Args:
            xuids: The Xbox Live IDs of the players.

        Parsers:
            - [User][spnkr.parsers.pydantic.profile.User]
            - [parse_users][spnkr.parsers.records.profile.parse_users]

        Returns:
            A list of users.
        """
        url = f"{_HOST}/users"
        params = {"xuids": [unwrap_xuid(x) for x in xuids]}
        return await self._get(url, params=params)



"""Game content management data services."""

from aiohttp import ClientResponse

from .base import BaseService

_HOST = "https://gamecms-hacs.svc.halowaypoint.com"


class GameCmsHacsService(BaseService):
    """Game content management data services."""

    async def get_medal_metadata(self) -> ClientResponse:
        """Get details for all medals obtainable in the game.

        Parsers:
            - [MedalMetadata][spnkr.parsers.pydantic.gamecms_hacs.MedalMetadata]
            - [parse_medal_metadata][spnkr.parsers.records.gamecms_hacs.parse_medal_metadata]

        Returns:
            The medal metadata.
        """
        url = f"{_HOST}/hi/Waypoint/file/medals/metadata.json"
        return await self._get(url)



"""Test SkillService."""

import pytest

from spnkr.services.skill import SkillService


@pytest.fixture
def service(session):
    return SkillService(session)


@pytest.mark.asyncio
async def test_get_match_skill(session, service: SkillService):
    await service.get_match_skill("match_id", [1234567890123456])
    session.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/matches/match_id/skill",
        params={"players": ["xuid(1234567890123456)"]},
    )


@pytest.mark.asyncio
async def test_get_playlist_csr(session, service: SkillService):
    await service.get_playlist_csr(
        "playlist_id", [1234567890123456, 2345678901234567]
    )
    session.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/playlist/playlist_id/csrs",
        params={
            "players": ["xuid(1234567890123456)", "xuid(2345678901234567)"]
        },
    )



"""Test GameContentHacsService."""

import pytest

from spnkr.services.gamecms_hacs import GameCmsHacsService


@pytest.fixture
def service(session):
    return GameCmsHacsService(session)


@pytest.mark.asyncio
async def test_get_medal_metadata(session, service: GameCmsHacsService):
    await service.get_medal_metadata()
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/Waypoint/file/medals/metadata.json"
    )



"""Test ProfileService."""

import pytest

from spnkr.services.profile import ProfileService


@pytest.fixture
def service(session):
    return ProfileService(session)


@pytest.mark.asyncio
async def test_get_current_user(session, service: ProfileService):
    await service.get_current_user()
    session.get.assert_called_with(
        "https://profile.svc.halowaypoint.com/users/me",
    )


@pytest.mark.asyncio
async def test_get_user_by_gamertag(session, service: ProfileService):
    await service.get_user_by_gamertag("MyGamertag")
    session.get.assert_called_with(
        "https://profile.svc.halowaypoint.com/users/gt(MyGamertag)",
    )


@pytest.mark.asyncio
async def test_get_user_by_id(session, service: ProfileService):
    await service.get_user_by_id("xuid(2345678901234567)")
    session.get.assert_called_with(
        "https://profile.svc.halowaypoint.com/users/xuid(2345678901234567)",
    )


@pytest.mark.asyncio
async def test_get_users_by_id(session, service: ProfileService):
    await service.get_users_by_id(
        ["xuid(1234567890123456)", "xuid(2345678901234567)"]
    )
    session.get.assert_called_with(
        "https://profile.svc.halowaypoint.com/users",
        params={"xuids": [1234567890123456, 2345678901234567]},
    )



"""Test BaseService."""

import time

import pytest

from spnkr.services.base import BaseService


@pytest.fixture
def service(session):
    return BaseService(session)


def test_async_limiter_set(service: BaseService):
    """Test that the async limiter is set as expected."""
    assert service._rate_limiter is not None
    assert (
        service._rate_limiter.max_rate / service._rate_limiter.time_period == 5
    )


@pytest.mark.asyncio
async def test_rate_limiter(service: BaseService):
    """Test that the rate limiter limits requests as expected."""
    t0 = time.time()
    # Use 6 requests due to the default rate of 5 requests per second.
    for _ in range(6):
        await service._get("url")
    t1 = time.time()
    assert t1 - t0 >= 1


@pytest.mark.asyncio
async def test_rate_limiter_none(service: BaseService):
    """Test that the rate limiter does not limit requests if None."""
    service._rate_limiter = None
    t0 = time.time()
    for _ in range(6):
        await service._get("url")
    t1 = time.time()
    assert t1 - t0 < 1



"""Parse flattened `NamedTuple` records from Halo Infinite API responses.

The record format can be useful for dumping to files or databases and is faster
than parsing the entirety of the JSON responses.
"""

from .gamecms_hacs import MedalRecord, parse_medal_metadata
from .profile import User, parse_user, parse_users
from .skill import (
    PlayerSkillRecord,
    PlaylistCsrRecord,
    parse_match_skill,
    parse_playlist_csr,
)
from .stats import (
    MatchCountRecord,
    MatchHistoryRecord,
    MatchInfoRecord,
    PlayerCoreStatsRecord,
    PlayerMedalRecord,
    PlayerServiceRecord,
    TeamCoreStatsRecord,
    parse_match_count,
    parse_match_history,
    parse_match_info,
    parse_player_core_stats,
    parse_player_medals,
    parse_service_record,
    parse_team_core_stats,
)
from .ugc_discovery import AssetRecord, parse_asset

__all__ = [
    "AssetRecord",
    "MatchCountRecord",
    "MatchHistoryRecord",
    "MatchInfoRecord",
    "MedalRecord",
    "PlayerCoreStatsRecord",
    "PlayerMedalRecord",
    "PlayerSkillRecord",
    "PlaylistCsrRecord",
    "PlayerServiceRecord",
    "TeamCoreStatsRecord",
    "User",
    "parse_asset",
    "parse_match_count",
    "parse_match_history",
    "parse_match_info",
    "parse_medal_metadata",
    "parse_player_core_stats",
    "parse_player_medals",
    "parse_match_skill",
    "parse_playlist_csr",
    "parse_service_record",
    "parse_team_core_stats",
    "parse_user",
    "parse_users",
]



import json
from pathlib import Path
from typing import Any
from uuid import UUID

import pytest

from spnkr.parsers import pydantic as pp

RESPONSES = Path("tests/responses")


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


def test_parse_medal_metadata():
    data = load_response("get_medal_metadata")
    result = pp.MedalMetadata(**data)
    assert result.medals[0].name_id == 622331684


def test_parse_match_skill():
    data = load_response("get_match_skill")
    result = pp.MatchSkill(**data)
    assert result.value[0].id == "xuid(2535445291321133)"


def test_parse_playlist_csr():
    data = load_response("get_playlist_csr")
    result = pp.PlaylistCsr(**data)
    assert result.value[0].id == "xuid(2535445291321133)"


def test_parse_match_count():
    data = load_response("get_match_count")
    result = pp.MatchCount(**data)
    assert result.matchmade_matches_played_count == 729


def test_parse_service_record():
    data = load_response("get_service_record")
    result = pp.ServiceRecord(**data)
    assert result.matches_completed == 1915


def test_parse_service_record_empty():
    data = load_response("get_service_record_empty")
    result = pp.ServiceRecord(**data)
    assert result.matches_completed == 0


def test_parse_match_history():
    data = load_response("get_match_history")
    result = pp.MatchHistory(**data)
    expected = UUID("a1219f5c-5942-4cd5-9ff1-4b6ea89905bc")
    assert result.results[0].match_id == expected


def test_parse_match_stats():
    data = load_response("get_match_stats")
    result = pp.MatchStats(**data)
    expected = UUID("6f050134-bede-47bc-a6df-eeafdcb9f97f")
    assert result.match_id == expected


def test_parse_match_stats_ctf():
    data = load_response("get_match_stats_ctf")
    result = pp.MatchStats(**data)
    expected = UUID("ffccbba6-fd52-4ff8-aa41-4bdc6487cd92")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.capture_the_flag_stats
    assert mode is not None


def test_parse_match_stats_zones():
    data = load_response("get_match_stats_zones")
    result = pp.MatchStats(**data)
    expected = UUID("ff9af871-ea6b-40c4-abe6-9aab6bfe1808")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.zones_stats
    assert mode is not None


def test_parse_match_stats_elimination():
    data = load_response("get_match_stats_elimination")
    result = pp.MatchStats(**data)
    expected = UUID("fe368e0b-9281-43ad-9b3d-9b16fe1e9402")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.elimination_stats
    assert mode is not None


def test_parse_match_stats_oddball():
    data = load_response("get_match_stats_oddball")
    result = pp.MatchStats(**data)
    expected = UUID("fe75a257-acd6-41ac-a778-8f86765639cb")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.oddball_stats
    assert mode is not None


def test_parse_match_stats_stockpile():
    data = load_response("get_match_stats_stockpile")
    result = pp.MatchStats(**data)
    expected = UUID("f40f4aa4-cc28-466d-b4ec-86fd86424083")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.stockpile_stats
    assert mode is not None


def test_parse_match_stats_extraction():
    data = load_response("get_match_stats_extraction")
    result = pp.MatchStats(**data)
    expected = UUID("44b9ca37-1d91-438a-8ef1-aa4cb78a19b2")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.extraction_stats
    assert mode is not None


def test_parse_game_variant():
    data = load_response("get_ugc_game_variant")
    result = pp.UgcGameVariant(**data)
    assert result.public_name == "Fiesta:Slayer"


def test_parse_map_mode_pair():
    data = load_response("get_map_mode_pair")
    result = pp.MapModePair(**data)
    assert result.public_name == "Arena:Shotty Snipes Slayer on Catalyst"


def test_parse_map():
    data = load_response("get_map")
    result = pp.Map(**data)
    assert result.public_name == "Forge Space"


def test_parse_playlist():
    data = load_response("get_playlist")
    result = pp.Playlist(**data)
    assert result.public_name == "Ranked Arena"


def test_parse_user():
    data = load_response("get_user")
    result = pp.User(**data)
    assert result.xuid == 1234567890


def test_parse_users():
    data = load_response("get_users")
    result = [pp.User(**user) for user in data]
    assert len(result) == 1
    assert result[0].xuid == 1234567890


if __name__ == "__main__":
    pytest.main()



"""Test the spnkr.auth.player module."""

from spnkr.auth import halo, player


def test_authenticated_player_is_valid_true():
    """Test that the authenticated player is valid."""
    expires_utc = {"ISO8601Date": "2999-09-01T00:00:00.000Z"}
    spartan_token = halo.SpartanToken({"ExpiresUtc": expires_utc})
    clearance_token = halo.ClearanceToken({"FlightConfigurationId": "token"})
    authenticated_player = player.AuthenticatedPlayer(
        "player_id",
        "gamertag",
        spartan_token,
        clearance_token,
        "xbl_auth_value",
    )
    assert authenticated_player.is_valid


def test_authenticated_player_is_valid_false():
    """Test that the authenticated player is not valid."""
    expires_utc = {"ISO8601Date": "1999-09-01T00:00:00.000Z"}
    spartan_token = halo.SpartanToken({"ExpiresUtc": expires_utc})
    clearance_token = halo.ClearanceToken({"FlightConfigurationId": "token"})
    authenticated_player = player.AuthenticatedPlayer(
        "player_id",
        "gamertag",
        spartan_token,
        clearance_token,
        "xbl_auth_value",
    )
    assert not authenticated_player.is_valid



"""Parsing functions for the "skill" authority."""

from typing import Any, NamedTuple
from uuid import UUID

from ..refdata import SkillResultCode


class PlayerSkillRecord(NamedTuple):
    """A player's skill data for a single match.

    Attributes:
        match_id: The match GUID.
        player_id: The player's Xbox Live ID.
        result_code: The status of the skill result.
        team_id: The player's team ID.
        team_mmr: The average MMR of the player's team.
        pre_match_csr: The player's CSR before the match.
        post_match_csr: The player's CSR after the match.
        actual_kills: The number of kills the player got in the match.
        expected_kills: The number of kills the player was expected to get in the match.
    """

    match_id: UUID
    player_id: str
    result_code: SkillResultCode
    team_id: int
    team_mmr: int
    pre_match_csr: int
    post_match_csr: int
    actual_kills: int
    expected_kills: int
    actual_deaths: int
    expected_deaths: int


class PlaylistCsrRecord(NamedTuple):
    """A player's CSR summary for a playlist.

    Attributes:
        player_id: The player's Xbox Live ID.
        result_code: The status of the CSR request.
        current_csr: The player's current CSR in the playlist.
        season_max_csr: The player's highest CSR in the playlist for the current season.
        all_time_max_csr: The player's highest CSR in the playlist for all seasons.
    """

    player_id: str
    result_code: SkillResultCode
    current_csr: int
    season_max_csr: int
    all_time_max_csr: int


def parse_match_skill(
    match_id: str | UUID, match_skill: dict[str, Any]
) -> list[PlayerSkillRecord]:
    """Parse a match skill response into a list of player skill records.

    Args:
        match_id: The match GUID.
        match_skill: The deserialized JSON from the client's `get_match_skill`
            method.

    Returns:
        A list of player skill records.
    """
    return [
        _parse_match_skill_value(match_id, msv) for msv in match_skill["Value"]
    ]


def _parse_match_skill_value(
    match_id: str | UUID, match_skill_value: dict[str, Any]
) -> PlayerSkillRecord:
    """Parse a single player's skill from the `get_match_skill` response JSON."""
    msv = match_skill_value
    result = msv["Result"]
    return PlayerSkillRecord(
        match_id=UUID(str(match_id)),
        player_id=msv["Id"],
        result_code=SkillResultCode(msv["ResultCode"]),
        team_id=result["TeamId"],
        team_mmr=result["TeamMmr"],
        pre_match_csr=result["RankRecap"]["PreMatchCsr"]["Value"],
        post_match_csr=result["RankRecap"]["PostMatchCsr"]["Value"],
        actual_kills=result["StatPerformances"]["Kills"]["Count"],
        expected_kills=result["StatPerformances"]["Kills"]["Expected"],
        actual_deaths=result["StatPerformances"]["Deaths"]["Count"],
        expected_deaths=result["StatPerformances"]["Deaths"]["Expected"],
    )


def parse_playlist_csr(playlist_csr: dict[str, Any]) -> list[PlaylistCsrRecord]:
    """Parse a playlist CSR response into a list of player-playlist CSR records.

    Args:
        playlist_csr: The deserialized JSON from the client's `get_playlist_csr`
            method.

    Returns:
        A list of player CSR summary records.
    """
    return [_parse_playlist_csr_value(pcv) for pcv in playlist_csr["Value"]]


def _parse_playlist_csr_value(
    playlist_csr_value: dict[str, Any]
) -> PlaylistCsrRecord:
    """Parse a single player's CSR from the `get_playlist_csr` response JSON."""
    pcv = playlist_csr_value
    result = pcv["Result"]
    return PlaylistCsrRecord(
        player_id=pcv["Id"],
        result_code=SkillResultCode(pcv["ResultCode"]),
        current_csr=result["Current"]["Value"],
        season_max_csr=result["SeasonMax"]["Value"],
        all_time_max_csr=result["AllTimeMax"]["Value"],
    )



"""Test DiscoveryUgcService."""

import pytest

from spnkr.services.discovery_ugc import DiscoveryUgcService


@pytest.fixture
def service(session):
    return DiscoveryUgcService(session)


@pytest.mark.asyncio
async def test_get_map_mode_pair(session, service: DiscoveryUgcService):
    await service.get_map_mode_pair("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/mapModePairs/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_playlist(session, service: DiscoveryUgcService):
    await service.get_playlist("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/playlists/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_map(session, service: DiscoveryUgcService):
    await service.get_map("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/maps/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_ugc_game_variant(session, service: DiscoveryUgcService):
    await service.get_ugc_game_variant("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/ugcGameVariants/asset_id/versions/version_id"
    )



"""Parsing functions for the "gamecms_hacs" authority."""

from typing import Any, NamedTuple

from ..refdata import MedalDifficulty, MedalType


class MedalRecord(NamedTuple):
    """Details about a medal.

    Attributes:
        name_id: The medal's ID.
        name: The medal's name.
        description: The medal's description.
        sprite_index: The index of the medal's sprite in the sprite sheet.
        sorting_weight: The medal's sorting weight.
        difficulty: The medal's difficulty.
        type: The medal's type.
        personal_score: The amount of personal score awarded by obtaining the medal.
    """

    name_id: int
    name: str
    description: str
    sprite_index: int
    sorting_weight: int
    difficulty: MedalDifficulty
    type: MedalType
    personal_score: int


def parse_medal_metadata(data: dict[str, Any]) -> list[MedalRecord]:
    """Parse `get_medal_metadata` response JSON from the client.

    Args:
        data: The deserialized JSON from the client's `get_medal_metadata`
            method.

    Returns:
        A list of medal detail records.
    """
    return [_parse_medal(medal_data) for medal_data in data["medals"]]


def _parse_medal(medal_data: dict[str, Any]) -> MedalRecord:
    """Parse a single medal from the `get_medal_metadata` response JSON."""
    return MedalRecord(
        name_id=medal_data["nameId"],
        name=medal_data["name"]["value"],
        description=medal_data["description"]["value"],
        sprite_index=medal_data["spriteIndex"],
        sorting_weight=medal_data["sortingWeight"],
        difficulty=MedalDifficulty(medal_data["difficultyIndex"]),
        type=MedalType(medal_data["typeIndex"]),
        personal_score=medal_data["personalScore"],
    )



"""Test the SPNKR API client."""

import asyncio
import time

import pytest

from spnkr.client import HaloInfiniteClient
from spnkr.services import (
    DiscoveryUgcService,
    GameCmsHacsService,
    ProfileService,
    SkillService,
    StatsService,
)


@pytest.fixture
def client(session):
    return HaloInfiniteClient(session, "spartan", "clearance", 5)


def test_client_header_update(client: HaloInfiniteClient):
    """Test that the client headers are updated as expected."""
    assert client._session.headers == {
        "Accept": "application/json",
        "x-343-authorization-spartan": "spartan",
        "343-clearance": "clearance",
    }


def test_client_services(client: HaloInfiniteClient):
    """Test that the client services are created as expected."""
    assert isinstance(client.discovery_ugc, DiscoveryUgcService)
    assert isinstance(client.gamecms_hacs, GameCmsHacsService)
    assert isinstance(client.profile, ProfileService)
    assert isinstance(client.skill, SkillService)
    assert isinstance(client.stats, StatsService)


@pytest.mark.asyncio
async def test_client_requests_per_second_multiple_services(
    client: HaloInfiniteClient,
):
    """Test that requests per second is applied per service."""
    tasks = []
    for _ in range(3):
        tasks.append(client.discovery_ugc._get(""))
        tasks.append(client.gamecms_hacs._get(""))
        tasks.append(client.profile._get(""))
        tasks.append(client.skill._get(""))
        tasks.append(client.stats._get(""))
    t0 = time.time()
    await asyncio.gather(*tasks)
    t1 = time.time()
    assert t1 - t0 <= 1



"""Parsing functions for the "profile" authority."""

from typing import NamedTuple


class User(NamedTuple):
    """High-level information about an Xbox Live user.

    Attributes:
        xuid: Xbox user ID.
        gamertag: User's gamertag.
        gamerpic: URL to the user's gamerpic (1080x1080 PNG). The URL will
            accept additional query parameters `w` and `h` to request a smaller
            image having pixel width and height dimensions, respectively. Known
            sizes are 64x64, 208x208, and 424x424. For example, "&w=424&h=424".
    """

    xuid: int
    gamertag: str
    gamerpic: str


def parse_users(data: list[dict]) -> list[User]:
    """Parse a list of users from the "get_users" client response.

    Args:
        data: A "get_users" response from the client.

    Returns:
        List of parsed users.
    """
    return [parse_user(user) for user in data]


def parse_user(data: dict) -> User:
    """Parse a single user.

    Args:
        data: The user data.

    Returns:
        Parsed user.
    """
    return User(
        xuid=int(data["xuid"]),
        gamertag=data["gamertag"],
        gamerpic=data["gamerpic"]["xlarge"],
    )



"""Test record parsing of Halo Infinite API responses."""

import datetime as dt
import json
import uuid
from pathlib import Path
from typing import Any

import pytest

from spnkr.parsers import records
from spnkr.parsers.records.stats import _parse_iso_duration

RESPONSES = Path("tests/responses")
TEST_ISO8601_DURATIONS = [
    (dt.timedelta(seconds=0), "PT0S"),
    (dt.timedelta(seconds=1), "PT1S"),
    (dt.timedelta(seconds=60), "PT1M"),
    (dt.timedelta(seconds=3600), "PT1H"),
    (dt.timedelta(minutes=12, seconds=34.1102934), "PT12M34.1102934S"),
    (dt.timedelta(hours=1, seconds=1), "PT1H1S"),
    (
        dt.timedelta(days=12, hours=9, minutes=47, seconds=30.0357698),
        "P12DT9H47M30.0357698S",
    ),
]


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


@pytest.mark.parametrize("delta, iso", TEST_ISO8601_DURATIONS)
def test_parse_iso_duration(delta: dt.timedelta, iso: str):
    result = _parse_iso_duration(iso)
    assert result == delta


def test_parse_medal_metadata():
    data = load_response("get_medal_metadata")
    result = records.parse_medal_metadata(data)
    assert result[0].name_id == 622331684


def test_parse_match_skill():
    data = load_response("get_match_skill")
    result = records.parse_match_skill(uuid.uuid4(), data)
    assert result[0].player_id == "xuid(2535445291321133)"


def test_parse_playlist_csr():
    data = load_response("get_playlist_csr")
    result = records.parse_playlist_csr(data)
    assert result[0].player_id == "xuid(2535445291321133)"


def test_parse_match_count():
    data = load_response("get_match_count")
    result = records.parse_match_count("xuid(1234567890123456)", data)
    assert result.matchmade == 729


def test_parse_match_history():
    data = load_response("get_match_history")
    result = records.parse_match_history(data)
    expected = "a1219f5c-5942-4cd5-9ff1-4b6ea89905bc"
    assert result[0].match_id == expected


def test_parse_service_record():
    xuid = "xuid(2535445291321133)"
    data = load_response("get_service_record")
    result = records.parse_service_record(xuid, data)
    assert result.player_id == xuid


def test_parse_service_record_empty():
    xuid = "xuid(2535445291321133)"
    data = load_response("get_service_record_empty")
    result = records.parse_service_record(xuid, data)
    assert result.player_id == xuid


def test_parse_match_stats_match_info():
    data = load_response("get_match_stats")
    result = records.parse_match_info(data)
    expected = "6f050134-bede-47bc-a6df-eeafdcb9f97f"
    assert result.match_id == expected


def test_parse_match_stats_team_core_stats():
    data = load_response("get_match_stats")
    result = records.parse_team_core_stats(data)
    expected = "6f050134-bede-47bc-a6df-eeafdcb9f97f"
    assert result[0].match_id == expected


def test_parse_match_stats_player_core_stats():
    data = load_response("get_match_stats")
    result = records.parse_player_core_stats(data)
    expected = "6f050134-bede-47bc-a6df-eeafdcb9f97f"
    assert result[0].match_id == expected


def test_parse_match_stats_player_medal_stats():
    data = load_response("get_match_stats")
    result = records.parse_player_medals(data)
    expected = "6f050134-bede-47bc-a6df-eeafdcb9f97f"
    assert result[0].match_id == expected


def test_parse_game_variant():
    data = load_response("get_ugc_game_variant")
    result = records.parse_asset(data)
    assert result.name == "Fiesta:Slayer"


def test_parse_map_mode_pair():
    data = load_response("get_map_mode_pair")
    result = records.parse_asset(data)
    assert result.name == "Arena:Shotty Snipes Slayer on Catalyst"


def test_parse_map():
    data = load_response("get_map")
    result = records.parse_asset(data)
    assert result.name == "Forge Space"


def test_parse_playlist():
    data = load_response("get_playlist")
    result = records.parse_asset(data)
    assert result.name == "Ranked Arena"


def test_parse_user():
    data = load_response("get_user")
    result = records.parse_user(data)
    assert result.xuid == 1234567890


def test_parse_users():
    data = load_response("get_users")
    result = records.parse_users(data)
    assert len(result) == 1
    assert result[0].xuid == 1234567890


if __name__ == "__main__":
    pytest.main()



"""Parsing functions for the "ugc_discovery" authority."""

from typing import Any, NamedTuple


class AssetRecord(NamedTuple):
    """A game asset, such as a map or game mode.

    Attributes:
        name: The asset's name.
        description: The asset's description.
        asset_id: The asset's GUID.
        version_id: The asset's version GUID.
        plays_recent: The number of times the asset has been played recently.
        plays_all_time: The number of times the asset has been played in total.
        average_rating: The asset's average rating.
        number_of_ratings: The number of ratings the asset has received.
    """

    name: str
    description: str
    asset_id: str
    version_id: str
    plays_recent: int
    plays_all_time: int
    average_rating: float
    number_of_ratings: int


def parse_asset(asset: dict[str, Any]) -> AssetRecord:
    """Parse game asset information from Halo Infinite API responses."""
    stats = asset["AssetStats"]
    return AssetRecord(
        name=asset["PublicName"],
        description=asset["Description"],
        asset_id=asset["AssetId"],
        version_id=asset["VersionId"],
        plays_recent=stats["PlaysRecent"],
        plays_all_time=stats["PlaysAllTime"],
        average_rating=stats["AverageRating"],
        number_of_ratings=stats["NumberOfRatings"],
    )



"""Base service class."""

from aiohttp import ClientResponse, ClientSession
from aiolimiter import AsyncLimiter


def _create_limiter(rate_per_second: int) -> AsyncLimiter:
    """Return an AsyncLimiter with the given rate per second."""
    # Setting the max rate to 1 disallows bursts.
    return AsyncLimiter(1, 1 / rate_per_second)


class BaseService:
    """Base service class. Handles initialization and rate limiting requests."""

    def __init__(
        self, session: ClientSession, requests_per_second: int | None = 5
    ) -> None:
        """Initialize a service.

        Args:
            session: The authenticated aiohttp session to use.
            requests_per_second: The rate limit to use. Set to None to disable
                rate limiting.
        """
        self._session = session
        self._rate_limiter = None
        if requests_per_second is not None:
            self._rate_limiter = _create_limiter(requests_per_second)

    async def _get(self, url: str, **kwargs) -> ClientResponse:
        """Make a GET request to the given URL."""
        if self._rate_limiter is None:
            return await self._session.get(url, **kwargs)
        async with self._rate_limiter:
            return await self._session.get(url, **kwargs)



"""Test the spnkr.xuid module."""

import pytest

from spnkr.errors import InvalidXuidError
from spnkr.xuid import unwrap_xuid, wrap_xuid, wrap_xuid_or_gamertag

WRAPPABLE = [
    ("1234567890123456", "xuid(1234567890123456)"),
    ("xuid(1234567890123456)", "xuid(1234567890123456)"),
    (1234567890123456, "xuid(1234567890123456)"),
]
UNWRAPPABLE = [
    ("1234567890123456", 1234567890123456),
    ("xuid(1234567890123456)", 1234567890123456),
    (1234567890123456, 1234567890123456),
]
INVALID = [
    123456789012345,
    "123456789012345",
    12345678901234567,
    "12345678901234567",
    "1234y567890",
    "(123)",
    "xuid(1234567890x)",
    "bid(1234567890)",
    "",
    " ",
    "xuid()",
    "xuid(1234567890123456]",
]


@pytest.mark.parametrize("value, expected", WRAPPABLE)
def test_wrap_xuid(value: str | int, expected: str):
    """Test that the XUID is wrapped."""
    assert wrap_xuid(value) == expected


@pytest.mark.parametrize("value, expected", UNWRAPPABLE)
def test_unwrap_xuid(value: str | int, expected: int):
    """Test that the XUID is unwrapped."""
    assert unwrap_xuid(value) == expected


@pytest.mark.parametrize("value", INVALID)
def test_unwrap_xuid_invalid(value: str | int):
    """Test that invalid XUIDs raise an error."""
    with pytest.raises(InvalidXuidError):
        unwrap_xuid(value)


def test_wrap_xuid_or_gamertag():
    """Test that gamertags are not wrapped."""
    assert wrap_xuid_or_gamertag(" MyGamertag ") == "MyGamertag"


def test_wrap_xuid_or_gamertag_invalid_int():
    """Test that invalid XUIDs raise an error."""
    with pytest.raises(InvalidXuidError):
        wrap_xuid_or_gamertag(123456789012345)


def test_wrap_xuid_or_gamertag_white_space():
    """Test that white space strings raise an error."""
    with pytest.raises(InvalidXuidError):
        wrap_xuid_or_gamertag(" ")



"""Test the spnkr.parsers.refdata module."""

import pytest

from spnkr.parsers.refdata import SubTier

SUBTIER_VALUES = [
    (SubTier.ONE, 1),
    (SubTier.TWO, 2),
    (SubTier.THREE, 3),
    (SubTier.FOUR, 4),
    (SubTier.FIVE, 5),
    (SubTier.SIX, 6),
]


@pytest.mark.parametrize("subtier, value", SUBTIER_VALUES)
def test_subtier_from_int(subtier: SubTier, value: int):
    """Test the from_int method of the SubTier class."""
    assert SubTier.from_int(value) == subtier


@pytest.mark.parametrize("subtier, value", SUBTIER_VALUES)
def test_subtier_to_int(subtier: SubTier, value: int):
    """Test the as_int method of the SubTier class."""
    assert subtier.to_int() == value



"""Models for the HIUGC_Discovery authority."""

from typing import Any
from uuid import UUID

from .base import PascalCaseModel


class OnlineUriReference(PascalCaseModel):
    """A reference to an online resource.

    Attributes:
        authority_id: The ID of the authority that hosts the resource.
        path: The path to the resource, relative to the authority's host.
        query_string: The query string of the resource's URL.
        retry_policy_id: The ID of the retry policy to use when accessing the resource.
        topic_name: The name of the topic to use when accessing the resource.
        acknowledgement_type_id: The ID of the acknowledgement type to use when accessing the resource.
        authentication_lifetime_extension_supported: Whether the resource supports extending the lifetime of an authentication token.
        clearance_aware: Whether the resource is clearance aware.
    """

    authority_id: str
    path: str
    query_string: str | None
    retry_policy_id: str
    topic_name: str
    acknowledgement_type_id: int
    authentication_lifetime_extension_supported: bool
    clearance_aware: bool


class AssetFiles(PascalCaseModel):
    """Information about the files that make up an asset.

    Attributes:
        prefix: The root path to the asset's files.
        file_relative_paths: The relative paths to the asset's files.
        prefix_endpoint: Details about the endpoint of `prefix`.
    """

    prefix: str
    file_relative_paths: list[str]
    prefix_endpoint: OnlineUriReference


class AssetStats(PascalCaseModel):
    """Statistics about an asset.

    Attributes:
        plays_recent: The number of times the asset has been played recently.
        plays_all_time: The number of times the asset has been played in total.
        favorites: The number of times the asset has been favorited.
        likes: The number of times the asset has been liked.
        bookmarks: The number of times the asset has been bookmarked.
        parent_asset_count: The number of assets that link to this asset.
        average_rating: The asset's average rating.
        number_of_ratings: The number of ratings the asset has received.
    """

    plays_recent: int
    plays_all_time: int
    favorites: int
    likes: int
    bookmarks: int
    parent_asset_count: int
    average_rating: float | None
    number_of_ratings: int


class Asset(PascalCaseModel):
    """A game asset, such as a map or game mode.

    Attributes:
        asset_id: The asset's GUID.
        version_id: The asset version's GUID.
        public_name: The asset's name.
        description: The asset's description.
        files: Information about the files that make up the asset.
        contributors: The names of the asset's contributors.
        asset_home: The asset's home.
        asset_stats: Statistics about the asset.
        inspection_result: The asset's inspection result.
        clone_behavior: The asset's clone behavior.
        order: The asset's order.
    """

    asset_id: UUID
    version_id: UUID
    public_name: str
    description: str
    files: AssetFiles
    contributors: list[str]
    asset_home: int
    asset_stats: AssetStats
    inspection_result: int
    clone_behavior: int
    order: int


class MapCustomData(PascalCaseModel):
    """Custom data related to map variant assets.

    Attributes:
        num_of_objects_on_map: The number of objects on the map.
        tag_level_id: The ID of the tag level.
        is_baked: Whether the map is baked.
        has_node_graph: Whether the map has a node graph.
    """

    num_of_objects_on_map: int
    tag_level_id: int
    is_baked: bool
    has_node_graph: bool


class Map(Asset):
    """A map variant asset.

    Attributes:
        custom_data: Custom data related to the map variant.
        tags: Keywords for the map variant asset, such as "343i".
        prefab_links: Details about prefabricated assets used by the map variant.
    """

    custom_data: MapCustomData
    tags: list[str]
    prefab_links: list[Asset] | None


class RotationWeight(PascalCaseModel):
    """A weighting for a map-mode pair in a playlist.

    Attributes:
        weight: The weight of the map-mode pair in the playlist.
    """

    weight: float


class RotationEntry(Asset):
    """A map-mode pair in a playlist.

    Attributes:
        metadata: A weighting for the map-mode pair in the playlist.
    """

    metadata: RotationWeight


class PlaylistEntry(PascalCaseModel):
    """A map-mode pair in a playlist.

    Attributes:
        map_mode_pair_asset_id: The UUID of the map-mode pair asset.
        metadata: A weighting for the map-mode pair in the playlist.
    """

    map_mode_pair_asset_id: str
    metadata: RotationWeight


class PlaylistCustomData(PascalCaseModel):
    """Custom data related to playlist assets.

    Attributes:
        playlist_entries: The map-mode pairs in the playlist.
        strategy: The playlist's strategy ID.
        min_teams: The minimum number of teams needed to make a match.
        min_team_size: The minimum number of team members needed to make a match.
        max_teams: The maximum number of teams allowed in a match.
        max_team_size: The maximum number of team members allowed in a match.
        max_team_imbalance: The allowable difference in team size.
        max_splitscreen_players_allowed: The maximum number of players allowed to play splitscreen.
        allow_friend_join_in_progress: Whether match players' friends are allowed to join a match in progress.
        allow_matchmaking_join_in_progress: Whether new players are allowed to join a match in progress.
        allow_bot_join_in_progress: Whether bots are allowed to join a match in progress.
        exit_experience_duration_sec: The duration of the exit experience, in seconds.
        fireteam_leader_kick_allowed: Whether fireteam leaders are allowed to kick members.
        disable_midgame_chat: Whether chat is disabled in game.
        allowed_device_inputs: The input device IDs allowed in the playlist (KB/M, controller, etc.).
        bot_difficulty: The difficulty of bots in the playlist.
        min_fireteam_size: The minimum number of players expected in a fireteam.
        max_fireteam_size: The maximum number of players allowed in a fireteam.
    """

    playlist_entries: list[PlaylistEntry]
    strategy: int
    min_teams: int
    min_team_size: int
    max_teams: int
    max_team_size: int
    max_team_imbalance: int
    max_splitscreen_players_allowed: int
    allow_friend_join_in_progress: bool
    allow_matchmaking_join_in_progress: bool
    allow_bot_join_in_progress: bool
    exit_experience_duration_sec: int
    fireteam_leader_kick_allowed: bool
    disable_midgame_chat: bool
    allowed_device_inputs: list[int]
    bot_difficulty: int
    min_fireteam_size: int
    max_fireteam_size: int


class Playlist(Asset):
    """A playlist asset.

    Attributes:
        custom_data: Custom data related to the playlist.
        tags: Keywords for the playlist asset, such as "343i".
        rotation_entries: The map-mode pairs in the playlist.
    """

    custom_data: PlaylistCustomData
    tags: list[str]
    rotation_entries: list[RotationEntry]


class MapModePair(Asset):
    """A map-mode pair asset.

    Attributes:
        custom_data: Custom data related to the map-mode pair.
        tags: Keywords for the map-mode pair asset, such as "343i".
        map_link: Details about the map variant asset.
        ugc_game_variant_link: Details about the game variant asset.
    """

    custom_data: dict[Any, Any]
    tags: list[str]
    map_link: Asset | None
    ugc_game_variant_link: Asset


class UgcGameVariantCustomData(PascalCaseModel):
    """Custom data related to game variant assets.

    Attributes:
        key_values: Key-value pairs of custom data.
    """

    key_values: dict[Any, Any]


class UgcGameVariant(Asset):
    """A user-generated game variant asset.

    Attributes:
        custom_data: Custom data related to the game variant.
        tags: Keywords for the game variant asset, such as "343i".
        engine_game_variant_link: Details about the mode's engine game variant.
    """

    custom_data: UgcGameVariantCustomData
    tags: list[str]
    engine_game_variant_link: Asset



"""Services available for retrieving Halo Infinite data.

Access service instances via
[HaloInfiniteClient][spnkr.client.HaloInfiniteClient] properties.
"""

from .discovery_ugc import DiscoveryUgcService
from .gamecms_hacs import GameCmsHacsService
from .profile import ProfileService
from .skill import SkillService
from .stats import StatsService

__all__ = [
    "DiscoveryUgcService",
    "GameCmsHacsService",
    "ProfileService",
    "SkillService",
    "StatsService",
]



"""Logic for obtaining API credentials."""

from .app import AzureApp
from .core import authenticate_player, refresh_player_tokens
from .halo import ClearanceToken, SpartanToken
from .player import AuthenticatedPlayer

__all__ = [
    "AzureApp",
    "AuthenticatedPlayer",
    "ClearanceToken",
    "SpartanToken",
    "authenticate_player",
    "refresh_player_tokens",
]



"""Test the spnkr.auth.core module."""

import asyncio

import pytest
from aiohttp import ClientSession

from spnkr.auth import app, core, halo, oauth, xbox


@pytest.fixture
def azure_app():
    return app.AzureApp("test_client_id", "test_client_secret")


@pytest.fixture
async def session():
    async with ClientSession() as session:
        yield session


def create_future(result):
    future = asyncio.Future()
    future.set_result(result)
    return future


class MockToken:
    def __init__(self, **kwargs):
        self.attrs = kwargs

    def __getattr__(self, name):
        return self.attrs[name]


@pytest.mark.asyncio
async def test_authenticate_player(session, azure_app, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: " test_code ")
    token = oauth.OAuth2Token({"refresh_token": "test_refresh_token"})
    future = create_future(token)
    monkeypatch.setattr(oauth, "request_oauth_token", lambda *args: future)
    refresh_token = await core.authenticate_player(session, azure_app)
    assert refresh_token == "test_refresh_token"


def test_get_authorization_code(azure_app, monkeypatch):
    """Test that the user-inputted code is returned."""
    monkeypatch.setattr("builtins.input", lambda _: " test_code ")
    assert core._get_authorization_code(azure_app) == "test_code"


@pytest.mark.asyncio
async def test_refresh_player_tokens(session, azure_app, monkeypatch):
    monkeypatch.setattr(
        oauth,
        "refresh_oauth_token",
        lambda *a: create_future(MockToken(access_token="test1")),
    )
    monkeypatch.setattr(
        xbox,
        "request_user_token",
        lambda *a: create_future(MockToken(token="test2")),
    )
    monkeypatch.setattr(
        xbox,
        "request_xsts_token",
        lambda *a: create_future(
            MockToken(
                token="test3",
                xuid="xuid(123)",
                gamertag="MrChief",
                authorization_header_value="X 123;ABC",
            )
        ),
    )
    monkeypatch.setattr(
        halo,
        "request_spartan_token",
        lambda *a: create_future(MockToken(token="test4")),
    )
    monkeypatch.setattr(
        halo,
        "request_clearance_token",
        lambda *a: create_future(MockToken(token="test5")),
    )
    player = await core.refresh_player_tokens(session, azure_app, "rt")
    assert player.player_id == "xuid(123)"
    assert player.gamertag == "MrChief"
    assert player.spartan_token.token == "test4"
    assert player.clearance_token.token == "test5"
    assert player.xbl_authorization_header_value == "X 123;ABC"
